"""Local GRPO finetuning implementation.

This module provides :class:`ArborGRPO`, a lightly adapted copy of DSPy's
``FinetuneTeleprompter``-based GRPO trainer so that Arbor can depend on the
implementation without pulling it directly from ``dspy`` at runtime. The code
is intentionally kept close to upstream to make future syncs straightforward.
"""

from __future__ import annotations

import logging
import random
import time
from collections import Counter, deque
from itertools import chain
from typing import Any, Callable, Literal

from dspy.adapters.base import Adapter
from dspy.adapters.chat_adapter import ChatAdapter
from dspy.adapters.xml_adapter import XMLAdapter
from dspy.clients.lm import LM
from dspy.clients.utils_finetune import GRPOGroup, GRPOStatus, TrainDataFormat
from dspy.dsp.utils.settings import settings
from dspy.evaluate.evaluate import Evaluate
from dspy.primitives.example import Example
from dspy.primitives.module import Module
from dspy.teleprompt.bootstrap_finetune import (
    FinetuneTeleprompter,
    all_predictors_have_lms,
    assert_structural_equivalency,
)
from dspy.teleprompt.bootstrap_trace import FailedPrediction, bootstrap_trace_data

logger = logging.getLogger(__name__)


class ArborGRPO(FinetuneTeleprompter):
    """GRPO finetuner for DSPy with Arbor defaults."""

    def __init__(
        self,
        metric: Callable | None = None,
        train_kwargs: dict[str, Any] | dict[LM, dict[str, Any]] | None = None,
        adapter: Adapter | dict[LM, Adapter] | None = None,
        exclude_demos: bool = False,
        num_threads: int = 6,
        num_train_steps: int = 500,
        seed: int = 0,
        num_dspy_examples_per_grpo_step: int = 1,
        num_rollouts_per_grpo_step: int = 4,
        use_train_as_val: bool = False,
        num_steps_for_val: int = 5,
        report_train_scores: bool = False,
        failure_score: float = 0,
        format_failure_score: float = -1,
        variably_invoked_predictor_grouping_mode: Literal["truncate"]
        | Literal["fill"]
        | Literal["ragged"] = "truncate",
        variably_invoked_predictor_fill_strategy: Literal["randint"]
        | Literal["max"]
        | None = None,
        checkpoint: Literal["single-best", "improvements", "none"] = "single-best",
    ):
        super().__init__(train_kwargs=train_kwargs)
        self.metric = metric
        self.adapter: dict[LM, Adapter] = self.convert_to_lm_dict(adapter)
        self.exclude_demos = exclude_demos
        self.num_threads = num_threads
        self.num_train_steps = num_train_steps
        self.rng = random.Random(seed)
        self.num_dspy_examples_per_grpo_step = num_dspy_examples_per_grpo_step
        assert num_rollouts_per_grpo_step > 1, (
            "num_rollouts_per_grpo_step must be greater than 1"
        )

        self.num_rollouts_per_grpo_step = num_rollouts_per_grpo_step
        self.use_train_as_val = use_train_as_val
        self.num_steps_for_val = num_steps_for_val
        self.report_train_scores = report_train_scores
        self.failure_score = failure_score
        self.format_failure_score = format_failure_score

        assert failure_score > format_failure_score, (
            "failure_score must be greater than format_failure_score since the range [format_failure_score, failure_score] is used to provide dspy formatting rewards"
        )

        if self.use_train_as_val:
            assert report_train_scores, (
                "If use_train_as_val is True, report_train_scores must be True."
            )

        assert exclude_demos, (
            "exclude_demos==False is not supported yet. Please set it to True."
        )

        self.variably_invoked_predictor_grouping_mode = (
            variably_invoked_predictor_grouping_mode
        )
        if variably_invoked_predictor_grouping_mode == "fill":
            assert variably_invoked_predictor_fill_strategy is not None, (
                "variably_invoked_predictor_fill_strategy must be set when variably_invoked_predictor_grouping_mode is 'fill'"
            )
            assert variably_invoked_predictor_fill_strategy in ["randint", "max"], (
                "variably_invoked_predictor_fill_strategy must be either 'randint' or 'max'"
            )
        self.variably_invoked_predictor_fill_strategy = (
            variably_invoked_predictor_fill_strategy
        )

        if checkpoint not in {"single-best", "improvements", "none"}:
            raise ValueError(
                "checkpoint must be one of 'single-best', 'improvements', or 'none'"
            )
        self.checkpoint_mode = checkpoint

        self.best_validation_score: float | None = None
        self.best_validation_step: int | None = None
        self.validation_scores: dict[int, float] = {}

        self.shuffled_trainset_ids = []
        self.epoch = -1
        self.id_freqs = Counter()
        self.fulfilled_batch_ids = []
        self.pending_batch_ids = []

    def validate_trace_data_and_log_issues(
        self,
        trace_data: list[list[list[dict[str, Any]]]],
        subsample_training_dataset: list[Example],
        num_teachers: int,
        num_samples_per_input: int,
        pred_signature_hash_to_ind: dict[int, int],
    ):
        assert len(trace_data) == len(subsample_training_dataset), (
            f"Trace data length {len(trace_data)} does not match the number of examples {len(subsample_training_dataset)}"
        )
        assert len(trace_data[0]) == num_teachers, (
            f"Trace data length {len(trace_data[0])} does not match the number of teachers {num_teachers}"
        )
        if len(trace_data[0][0]) == 0:
            logger.warning(
                "Trace data for example 0 and teacher 0 is empty. This is likely due to all examples in the training set input, resulting in the model generating output not following the dspy response format."
            )
        elif len(trace_data[0][0]) != num_samples_per_input:
            logger.warning(
                f"Trace data length {len(trace_data[0][0])} does not match the expected number of samples per input {num_samples_per_input}"
            )
            assert "trace" in trace_data[0][0][0], (
                "Trace data does not contain the 'trace' key"
            )
            assert len(trace_data[0][0][0]["trace"]) > 0, "Trace data is empty"
            assert len(trace_data[0][0][0]["trace"][0]) == 3, (
                f"Trace tuple length {len(trace_data[0][0][0]['trace'][0])} does not match the expected length 3"
            )

        for example_data in trace_data:
            for teacher_data in example_data:
                for sample in teacher_data:
                    for t in sample["trace"]:
                        assert hash(t[0].signature) in pred_signature_hash_to_ind

    def report_validation_metrics(
        self,
        student,
        trainset,
        valset,
        logger,
        step_idx: int = -1,
        grpo_training_job: Any | None = None,
        lm_for_job: LM | None = None,
    ):
        if (
            step_idx == -1
            or step_idx == self.num_train_steps - 1
            or (step_idx + 1) % self.num_steps_for_val == 0
        ):
            pass
        else:
            return

        score: float | None = None

        if valset is not None:
            assert not self.use_train_as_val, (
                "If valset is provided, use_train_as_val must be False."
            )
            assert (
                isinstance(self.num_steps_for_val, int) and self.num_steps_for_val > 0
            ), "num_steps_for_val must be a positive integer."
            if self.report_train_scores:
                if step_idx == -1:
                    logger.info(
                        "Using user provided validation set and reporting train scores for every validation step in addition."
                    )
                valset_evaluator = Evaluate(
                    devset=valset + trainset,
                    num_threads=self.num_threads,
                    display_progress=True,
                    provide_traceback=False,
                    max_errors=len(valset) * 10,
                    failure_score=self.failure_score,
                )
                if step_idx == -1:
                    logger.info(
                        "Evaluating the student program on the train+validation set before training loop..."
                    )
                else:
                    logger.info(
                        f"Evaluating the student program on the validation set after training step {step_idx + 1}/{self.num_train_steps}"
                    )
                valset_evaluation = valset_evaluator(student, metric=self.metric)
                trainset_scores = [
                    r[-1] for r in valset_evaluation.results[len(valset) :]
                ]
                valset_scores = [
                    r[-1] for r in valset_evaluation.results[: len(valset)]
                ]
                trainset_agg = sum(trainset_scores) / len(trainset_scores)
                valset_agg = sum(valset_scores) / len(valset_scores)
                score = valset_agg
                if step_idx == -1:
                    logger.info(
                        f"Student program training set score before training loop: {trainset_agg}"
                    )
                    logger.info(
                        f"Student program validation set score before training loop: {valset_agg}"
                    )
                else:
                    logger.info(
                        f"Student program training set score after training step {step_idx + 1}/{self.num_train_steps}: {trainset_agg}"
                    )
                    logger.info(
                        f"Student program validation set score after training step {step_idx + 1}/{self.num_train_steps}: {valset_agg}"
                    )
            else:
                if step_idx == -1:
                    logger.info(
                        "Using user provided validation set and not reporting train scores."
                    )
                valset_evaluator = Evaluate(
                    devset=valset,
                    num_threads=self.num_threads,
                    display_progress=True,
                    provide_traceback=False,
                    max_errors=len(valset) * 10,
                    failure_score=self.failure_score,
                )
                if step_idx == -1:
                    logger.info(
                        "Evaluating the student program on the validation set before training loop..."
                    )
                else:
                    logger.info(
                        f"Evaluating the student program on the validation set after training step {step_idx + 1}/{self.num_train_steps}"
                    )
                valset_evaluation = valset_evaluator(student, metric=self.metric)
                score = valset_evaluation.score
                if step_idx == -1:
                    logger.info(
                        f"Student program validation set score before training loop: {valset_evaluation.score}"
                    )
                else:
                    logger.info(
                        f"Student program validation set score after training step {step_idx + 1}/{self.num_train_steps}: {valset_evaluation.score}"
                    )
        else:
            if self.report_train_scores:
                assert self.use_train_as_val, (
                    "If report_train_scores is True, use_train_as_val must be True when valset is not provided explicitly."
                )
                assert (
                    isinstance(self.num_steps_for_val, int)
                    and self.num_steps_for_val > 0
                ), "num_steps_for_val must be a positive integer."
                if step_idx == -1:
                    logger.info("Using trainset as validation set.")
                valset_evaluator = Evaluate(
                    devset=trainset,
                    num_threads=self.num_threads,
                    display_progress=True,
                    provide_traceback=False,
                    max_errors=len(trainset) * 10,
                    failure_score=self.failure_score,
                )
                if step_idx == -1:
                    logger.info(
                        "Evaluating the student program on the validation set before training loop..."
                    )
                else:
                    logger.info(
                        f"Evaluating the student program on the validation set after training step {step_idx + 1}/{self.num_train_steps}"
                    )
                valset_evaluation = valset_evaluator(student, metric=self.metric)
                score = valset_evaluation.score
                if step_idx == -1:
                    logger.info(
                        f"Student program training set score before training loop: {valset_evaluation.score}"
                    )
                else:
                    logger.info(
                        f"Student program training set score after training step {step_idx + 1}/{self.num_train_steps}: {valset_evaluation.score}"
                    )
            else:
                assert not self.use_train_as_val, (
                    "If report_train_scores is False, use_train_as_val must be False."
                )
                if step_idx == -1:
                    logger.info(
                        "Not using any validation set and not reporting train scores."
                    )

        if score is None:
            return

        self.validation_scores[step_idx] = score

        if step_idx == -1:
            if self.best_validation_score is None or score > self.best_validation_score:
                self.best_validation_score = score
                self.best_validation_step = step_idx
            return

        improved = (
            self.best_validation_score is None or score > self.best_validation_score
        )

        if improved:
            logger.info(
                f"Saving checkpoint for step {step_idx + 1}/{self.num_train_steps}..."
            )
            self.best_validation_score = score
            self.best_validation_step = step_idx
            if (
                self.checkpoint_mode != "none"
                and grpo_training_job is not None
                and lm_for_job is not None
            ):
                checkpoint_name = (
                    "model_checkpoint_best"
                    if self.checkpoint_mode == "single-best"
                    else f"model_checkpoint_step_{step_idx + 1}"
                )
                self._save_checkpoint_for_job(
                    grpo_training_job,
                    lm_for_job,
                    checkpoint_name,
                    score=score,
                    step_idx=step_idx,
                )

    def update_shuffled_trainset(self, original_trainset):
        self.shuffled_trainset_ids = list(range(len(original_trainset)))
        self.rng.shuffle(self.shuffled_trainset_ids)
        for id in self.shuffled_trainset_ids:
            self.id_freqs[id] += 1

        num_to_pad = self.num_dspy_examples_per_grpo_step - (
            len(original_trainset) % self.num_dspy_examples_per_grpo_step
        )
        if num_to_pad > 0:
            for _ in range(num_to_pad):
                selected_id = self.id_freqs.most_common()[::-1][0][0]
                self.shuffled_trainset_ids.append(selected_id)
                self.id_freqs[selected_id] += 1

    def select_training_sample_and_update_shuffled_trainset(
        self,
        original_trainset: list[Example],
        train_step_idx: int,
    ) -> list[Example]:
        base_idx = train_step_idx * self.num_dspy_examples_per_grpo_step
        if self.epoch == -1:
            curr_epoch = 0
        else:
            curr_epoch = base_idx // len(self.shuffled_trainset_ids)
        if curr_epoch > self.epoch:
            logger.info(f"Updating shuffled trainset for epoch {curr_epoch}...")
            self.epoch = curr_epoch
            self.update_shuffled_trainset(original_trainset)

        assert len(self.shuffled_trainset_ids) >= self.num_dspy_examples_per_grpo_step
        assert (
            len(self.shuffled_trainset_ids) % self.num_dspy_examples_per_grpo_step == 0
        )

        base_idx = base_idx % len(self.shuffled_trainset_ids)
        end_idx = base_idx + self.num_dspy_examples_per_grpo_step
        assert end_idx <= len(self.shuffled_trainset_ids)
        selected_ids = self.shuffled_trainset_ids[base_idx:end_idx]
        selected_trainset = [original_trainset[i] for i in selected_ids]
        return selected_trainset

    def compile(
        self,
        student: Module,
        trainset: list[Example],
        teacher: Module | list[Module] | None = None,
        valset: list[Example] | None = None,
        **kwargs,
    ) -> Module:
        logger.info(
            "Starting the GRPO compilation process... The LM(s) for the student program will be updated in place at the end of the training."
        )
        logger.info("Validating the inputs...")

        assert len(trainset) > 0, (
            "Training set is empty. Please provide a non-empty training set."
        )

        if len(trainset) < self.num_dspy_examples_per_grpo_step:
            logger.warning(
                f"Number of training examples {len(trainset)} is less than the number of examples per GRPO step {self.num_dspy_examples_per_grpo_step}. "
                "Repeating the training set to fill the GRPO step. This could lead to overfitting and training instability."
            )
            multiplier = (
                self.num_dspy_examples_per_grpo_step + len(trainset) - 1
            ) // len(trainset)
            if multiplier > 1:
                logger.warning(
                    f"Repeating the training set {multiplier} times to fill the GRPO step. This could lead to overfitting and training instability."
                )
                trainset = trainset * multiplier

        student_lms = {id(pred.lm) for pred in student.predictors()}
        assert len(student_lms) == 1, (
            f"Student program has multiple LMs: {student_lms}. "
            "GRPO only supports student programs with a single LM."
            "You can set the LM for a program with `program.set_lm(...)`"
        )

        if self.use_train_as_val:
            assert valset is None, "If use_train_as_val is True, valset must be None."

        logger.info("Preparing the student program...")
        all_predictors_have_lms(student)
        pred_signature_hash_to_ind = {
            hash(pred.signature): ind for ind, pred in enumerate(student.predictors())
        }
        num_student_predictors = len(student.predictors())

        logging.info(
            "Preparing the teacher program(s)... We will ensure that the provided programs have the same program structure as the student program."
        )
        if (isinstance(teacher, list) and len(teacher) == 0) or teacher is None:
            teacher = student
        teachers = teacher if isinstance(teacher, list) else [teacher]
        for t in teachers:
            assert_structural_equivalency(student, t)
            all_predictors_have_lms(t)

        assert student in teachers, (
            f"Student program {student} is not in the list of teachers {teachers}. Please provide the student program as one of the teachers. Alternatively, you can leave the teacher argument as None, and the student program will be used as the teacher program."
        )
        assert self.num_rollouts_per_grpo_step % len(teachers) == 0, (
            f"The GRPO group size (num_rollouts_per_grpo_step) {self.num_rollouts_per_grpo_step} is not divisible by the number of teachers {len(teachers)}. "
            "This is required to ensure that each teacher gets the same number of examples."
            "Please provide a number of examples that is divisible by the number of teachers."
        )
        num_samples_per_input = self.num_rollouts_per_grpo_step // len(teachers)

        lm_cache_dict = {}
        disable_lm_cache(program=student, lm_cache_dict=lm_cache_dict)
        for t in teachers:
            disable_lm_cache(program=t, lm_cache_dict=lm_cache_dict)

        for pred in student.predictors():
            train_kwargs = self.train_kwargs[pred.lm]
            train_kwargs = {} if train_kwargs is None else train_kwargs
            train_kwargs["num_generations"] = self.num_rollouts_per_grpo_step
            self.train_kwargs[pred.lm] = train_kwargs

        logger.info("Preparing the GRPO training job...")
        student_predictors = list(student.predictors())
        assert student_predictors, "Student program must define at least one predictor."
        first_lm = student_predictors[0].lm
        for pred in student_predictors[1:]:
            assert pred.lm is first_lm, (
                "Multiple LMs detected for student predictors; only one GRPO training job is supported."
            )

        train_kwargs = self.train_kwargs[first_lm]
        # ensure the important sampling hyperparameters match between the train config and the DSPy lm sampling config
        for _hparam in ["temperature", "top_k", "top_p", "repetition_penalty"]:
            if _hparam not in train_kwargs:
                raise ValueError(
                    f"{_hparam} not found in train_kwargs. Please provide it in the train_kwargs."
                )
            if _hparam not in first_lm.kwargs:
                first_lm.kwargs[_hparam] = train_kwargs.get(_hparam)
                logger.warning(
                    f"Setting {_hparam} to {train_kwargs.get(_hparam)} for LM as it is not set in the train_kwargs."
                )
            elif _hparam in first_lm.kwargs and first_lm.kwargs.get(
                _hparam
            ) != train_kwargs.get(_hparam):
                raise ValueError(
                    f"{_hparam} mismatch for LM {first_lm}. Expected {train_kwargs.get(_hparam)}, got {first_lm.kwargs.get(_hparam)}"
                )

        grpo_training_job = first_lm.reinforce(train_kwargs=train_kwargs)

        group_queue: deque = deque()
        logger.info("Starting the GRPO training loop...")
        for train_step_idx in range(self.num_train_steps):
            logger.info(
                f"GRPO training step {train_step_idx + 1}/{self.num_train_steps}..."
            )

            subsample_training_dataset = (
                self.select_training_sample_and_update_shuffled_trainset(
                    original_trainset=trainset,
                    train_step_idx=train_step_idx,
                )
            )

            def _any_available_for_step():
                grpo_status: GRPOStatus = grpo_training_job.get_status()
                pending_batch_ids = grpo_status["pending_batch_ids"]
                available = set(pending_batch_ids) - set(self.fulfilled_batch_ids)
                return bool(available)

            while not _any_available_for_step():
                time.sleep(1)

            logger.info("Bootstrapping data...")
            format_failure_count = 0
            trace_data = [
                [[] for _ in range(len(teachers))]
                for _ in range(len(subsample_training_dataset))
            ]
            for tind, teacher in enumerate(teachers):
                subsample_training_dataset_repeated = [
                    example
                    for _ in range(num_samples_per_input)
                    for example in subsample_training_dataset
                ]
                round_data = bootstrap_trace_data(
                    program=teacher,
                    dataset=subsample_training_dataset_repeated,
                    metric=self.metric,
                    num_threads=self.num_threads,
                    raise_on_error=False,
                    capture_failed_parses=True,
                    failure_score=self.failure_score,
                    format_failure_score=self.format_failure_score,
                    log_format_failures=True,
                )
                for data_dict in round_data:
                    example_ind_in_subsample = data_dict["example_ind"] % len(
                        subsample_training_dataset
                    )
                    data_dict["example_ind"] = example_ind_in_subsample
                    trace_data[example_ind_in_subsample][tind].append(data_dict)

            self.validate_trace_data_and_log_issues(
                trace_data=trace_data,
                subsample_training_dataset=subsample_training_dataset,
                num_teachers=len(teachers),
                num_samples_per_input=num_samples_per_input,
                pred_signature_hash_to_ind=pred_signature_hash_to_ind,
            )

            logger.info(
                "Preparing the training data batch from bootstrapped examples for GRPO..."
            )
            train_batch_per_predictor: list[list[GRPOGroup]] = [
                [] for _ in range(num_student_predictors)
            ]
            for pred_id in range(num_student_predictors):
                for example_ind, example_data in enumerate(trace_data):
                    predictor_example_invocations: list[list[tuple]] = []

                    for teacher_data in example_data:
                        for sample in teacher_data:
                            assert sample["example_ind"] == example_ind

                            trace_instances_for_current_pred = [
                                (*t, sample["score"])
                                for t in sample["trace"]
                                if hash(t[0].signature)
                                == hash(student.predictors()[pred_id].signature)
                            ]

                            predictor_example_invocations.append(
                                trace_instances_for_current_pred
                            )

                    if len(predictor_example_invocations) == 0:
                        logger.warning(
                            f"Skipping example {example_ind} for predictor {pred_id} as it has no invocations. This is likely due to all examples in the training set input, resulting in the model generating output not following the dspy response format."
                        )
                        continue
                    elif (
                        len(predictor_example_invocations)
                        != self.num_rollouts_per_grpo_step
                    ):
                        logger.warning(
                            f"Number of predictor example invocations {len(predictor_example_invocations)} does not match the expected batch size {self.num_rollouts_per_grpo_step}. This is likely due to all examples in the training set input, resulting in the model generating output not following the dspy response format."
                        )

                    min_len = min(
                        [
                            len(predictor_example_invocations[i])
                            for i in range(len(predictor_example_invocations))
                        ]
                    )
                    max_len = max(
                        [
                            len(predictor_example_invocations[i])
                            for i in range(len(predictor_example_invocations))
                        ]
                    )
                    if min_len == 0:
                        logger.warning(
                            f"Skipping example {example_ind} for predictor {pred_id} as it has no invocations."
                        )
                        continue

                    if self.variably_invoked_predictor_grouping_mode == "truncate":
                        predictor_example_invocations = [
                            invocation[:min_len]
                            for invocation in predictor_example_invocations
                        ]
                    elif self.variably_invoked_predictor_grouping_mode == "fill":
                        if self.variably_invoked_predictor_fill_strategy == "randint":
                            selector = lambda l: self.rng.choice(l)  # noqa: E731, E741
                        else:
                            selector = lambda l: l[-1]  # noqa: E731, E741
                        predictor_example_invocations = [
                            invocation
                            + [
                                selector(invocation)
                                for _ in range(max_len - len(invocation))
                            ]
                            for invocation in predictor_example_invocations
                        ]
                    else:
                        assert self.variably_invoked_predictor_grouping_mode == "ragged"
                    max_len = max(
                        [
                            len(predictor_example_invocations[i])
                            for i in range(len(predictor_example_invocations))
                        ]
                    )

                    example_training_data: list[GRPOGroup] = [
                        [] for _ in range(max_len)
                    ]

                    for group_idx in range(max_len):
                        for rollout_idx in range(len(predictor_example_invocations)):
                            trace_instance = predictor_example_invocations[rollout_idx][
                                group_idx
                            ]
                            score = trace_instance[3]

                            trace_pred_id = pred_signature_hash_to_ind.get(
                                hash(trace_instance[0].signature)
                            )
                            assert trace_pred_id == pred_id

                            predictor = trace_instance[0]
                            pred_lm = predictor.lm
                            adapter = (
                                self.adapter[pred_lm]
                                or settings.adapter
                                or XMLAdapter()
                            )
                            assert isinstance(adapter, ChatAdapter), (
                                f"Adapter {adapter} is not a ChatAdapter. GRPO training is not supported for this adapter."
                            )
                            # Force it for now for testing purposes
                            adapter.use_json_adapter_fallback = False
                            assert not adapter.use_json_adapter_fallback, (
                                "Adapter fallback must be disabled for GRPO training."
                            )

                            inp_messages = adapter.format(
                                signature=trace_instance[0].signature,
                                inputs=trace_instance[1],
                                demos=[],
                            )

                            if isinstance(trace_instance[2], FailedPrediction):
                                score = (
                                    trace_instance[2].format_reward
                                    or self.format_failure_score
                                )
                                format_failure_count += 1
                                example_training_data[group_idx].append(
                                    {
                                        "messages": inp_messages,
                                        "completion": {
                                            "role": "assistant",
                                            "content": trace_instance[
                                                2
                                            ].completion_text,
                                        },
                                        "reward": float(score),
                                    }
                                )
                                logger.warning(
                                    f"Adding a format failure example to the training data for predictor {pred_id} and example {example_ind}."
                                )
                            else:
                                all_messages = adapter.format_finetune_data(
                                    signature=trace_instance[0].signature,
                                    inputs=trace_instance[1],
                                    outputs=trace_instance[2],
                                    demos=[],
                                )["messages"]

                                assert all_messages[:-1] == inp_messages

                                example_training_data[group_idx].append(
                                    {
                                        "messages": inp_messages,
                                        "completion": {
                                            "role": all_messages[-1]["role"],
                                            "content": all_messages[-1]["content"],
                                        },
                                        "reward": float(score),
                                    }
                                )

                    train_batch_per_predictor[pred_id].extend(example_training_data)

            if not any(train_batch_per_predictor):
                logger.warning(
                    "No training data found for this training step. This means that the model did not generate valid formatted responses for any of the examples in the training set. This is a critical error. Please check the model and the training set."
                )
                continue

            for predictor_train_batch in train_batch_per_predictor:
                for grpo_train_group in predictor_train_batch:
                    if len(grpo_train_group) != self.num_rollouts_per_grpo_step:
                        logger.warning(
                            f"Number of completions {len(grpo_train_group)} does not match the expected number num_rollouts_per_grpo_step={self.num_rollouts_per_grpo_step}"
                        )
                        assert (
                            len(grpo_train_group) <= self.num_rollouts_per_grpo_step
                        ), (
                            f"Number of completions {len(grpo_train_group)} is greater than the expected number num_rollouts_per_grpo_step={self.num_rollouts_per_grpo_step}"
                        )
                    if len(set(map(repr, grpo_train_group))) < 2:
                        logger.warning(
                            "GRPOGroup has no diversity. This could be due to low temperature, or low number of rollouts, or the cache could be enabled inadvertently. The GRPOGroup is %s.",
                            grpo_train_group,
                        )

            logger.info("Invoking GRPO training step...")
            all_train_data: list[GRPOGroup] = list(
                chain.from_iterable(train_batch_per_predictor)
            )
            train_data: list[GRPOGroup] = all_train_data
            for group in train_data:
                if len(group) != self.num_rollouts_per_grpo_step:
                    while len(group) < self.num_rollouts_per_grpo_step:
                        group.extend(
                            group[
                                : min(
                                    self.num_rollouts_per_grpo_step - len(group),
                                    len(group),
                                )
                            ]
                        )
                assert len(group) == self.num_rollouts_per_grpo_step, (
                    f"Number of completions {len(group)} does not match the expected number self.num_rollouts_per_grpo_step={self.num_rollouts_per_grpo_step}"
                )

            grpo_status: GRPOStatus = grpo_training_job.get_status()
            pending_batch_ids = grpo_status["pending_batch_ids"]
            available_batch_ids = list(
                set(pending_batch_ids) - set(self.fulfilled_batch_ids)
            )
            if not available_batch_ids:
                continue

            if len(group_queue) < len(available_batch_ids) and len(train_data) > 0:
                need = len(available_batch_ids) - len(group_queue)
                while need > 0:
                    shuffled = self.rng.sample(train_data, k=len(train_data))
                    group_queue.extend(shuffled)
                    need -= len(shuffled)

            # Report validation metrics if the step is the last step or a multiple of num_steps_for_val
            # Do this before submitting the train data to the GRPO training job so it can be attached to the train data
            if (
                train_step_idx == 0
                or train_step_idx == self.num_train_steps - 1
                or (train_step_idx + 1) % self.num_steps_for_val == 0
            ):
                self.report_validation_metrics(
                    student=student,
                    trainset=trainset,
                    valset=valset,
                    logger=logger,
                    step_idx=train_step_idx,
                    grpo_training_job=grpo_training_job,
                    lm_for_job=first_lm,
                )

            final_train_data: list[GRPOGroup] = []
            for bid in available_batch_ids:
                grp = group_queue.popleft()

                # Attach simple step metrics for logging on the backend
                step_metrics = {}
                if self.validation_scores.get(train_step_idx) is not None:
                    step_metrics["validation_score"] = float(
                        self.validation_scores[train_step_idx]
                    )
                step_metrics["format_failures"] = int(format_failure_count)
                final_train_data.append(
                    {"batch_id": bid, "group": grp, "metrics": step_metrics}
                )

            if not final_train_data:
                continue

            self.fulfilled_batch_ids.extend(
                [item["batch_id"] for item in final_train_data]
            )

            grpo_training_job.step(
                train_data=final_train_data,
                train_data_format=TrainDataFormat.GRPO_CHAT,
            )

            logger.info(
                f"GRPO training step {train_step_idx + 1}/{self.num_train_steps} completed."
            )

        logger.info("Done with the iterations! Retrieving the final model...")
        grpo_training_job.terminate()

        recover_lm_cache(program=student, lm_cache_dict=lm_cache_dict)
        for t in teachers:
            recover_lm_cache(program=t, lm_cache_dict=lm_cache_dict)

        logger.info("GRPO compiler has finished compiling the student program")
        student._compiled = True
        return student

    def _save_checkpoint_for_job(
        self,
        grpo_training_job: Any,
        lm_for_job: LM,
        checkpoint_name: str,
        *,
        score: float | None = None,
        step_idx: int | None = None,
    ) -> None:
        metadata = None
        if score is not None or step_idx is not None:
            metadata = {}
            if score is not None:
                metadata["score"] = float(score)
            if step_idx is not None:
                metadata["step_index"] = int(step_idx)

        kwargs: dict[str, Any] = {"checkpoint_name": checkpoint_name}
        if score is not None:
            kwargs["score"] = score
        if metadata:
            kwargs["metadata"] = metadata

        try:
            grpo_training_job.save_checkpoint(**kwargs)
        except TypeError:
            kwargs.pop("metadata", None)
            grpo_training_job.save_checkpoint(**kwargs)
        checkpoints = grpo_training_job.checkpoints or {}
        checkpoint_record = checkpoints.get(checkpoint_name, {})
        checkpoint_dir = checkpoint_record.get("checkpoint_dir")
        job_label = (
            lm_for_job.model
            if hasattr(lm_for_job, "model")
            else lm_for_job.name
            if hasattr(lm_for_job, "name")
            else lm_for_job.__class__.__name__
        )
        if checkpoint_dir:
            logger.info(
                "Saved checkpoint %s for %s at %s",
                checkpoint_name,
                job_label,
                checkpoint_dir,
            )
        else:
            logger.info("Saved checkpoint %s for %s", checkpoint_name, job_label)


def disable_lm_cache(program: Module, lm_cache_dict: dict):
    """Disable the LM cache for all predictors in the program."""

    for pred in program.predictors():
        if not pred.lm:
            raise ValueError(
                f"Cannot disable cache: predictor {pred} does not have an LM set."
            )
        if pred.lm not in lm_cache_dict:
            lm_cache_dict[pred.lm] = pred.lm.cache
        pred.lm.cache = False


def recover_lm_cache(program: Module, lm_cache_dict: dict):
    """Recover the LM caches for all predictors in the program to their original state."""

    for pred in program.predictors():
        if pred.lm in lm_cache_dict:
            pred.lm.cache = lm_cache_dict[pred.lm]
        else:
            pred.lm.cache = True


__all__ = ["ArborGRPO", "disable_lm_cache", "recover_lm_cache"]



from typing import Any, Optional

from arbor.core.config import Config
from arbor.server.api.schemas import InferenceLaunchRequest
from arbor.server.services.comms.control_server import TrainerControlServer
from arbor.server.services.jobs.inference_job import InferenceJob
from arbor.server.services.jobs.inference_launch_config import InferenceLaunchConfig
from arbor.server.services.managers.base_manager import BaseManager
from arbor.server.services.managers.gpu_manager import GPUManager


class InferenceManager(BaseManager):
    def __init__(self, config: Config, gpu_manager: GPUManager):
        super().__init__(config)
        self.inference_jobs: dict[str, InferenceJob] = {}
        self.alias_to_job_id: dict[str, str] = {}
        if gpu_manager is None:
            raise ValueError("InferenceManager requires a GPUManager instance")
        self.gpu_manager = gpu_manager

    # TODO: request_json should be checked for launch_model_config or something
    async def route_inference(self, request_json: dict):
        model = request_json["model"]
        self.logger.debug(f"Running inference for model {model}")

        # If model isnt launched, launch it
        # TODO: Check that there are GPUs available. If not, do hot swap or something.
        inference_job = self._get_job_by_alias(model)
        if inference_job is None:
            try:
                inference_job = self.launch_from_request(
                    InferenceLaunchRequest(model=model)
                )
            except Exception as e:
                self.logger.error(f"Error launching model {model}: {e}")
                raise e

        return await inference_job.run_inference(request_json)

    def _register_job(self, job: InferenceJob, aliases: list[str]) -> None:
        self.inference_jobs[job.id] = job
        for alias in aliases:
            if alias:
                self.alias_to_job_id[alias] = job.id
        self._cleanup_called = False

    def _get_job_by_alias(self, alias: Optional[str]) -> Optional[InferenceJob]:
        if not alias:
            return None
        job_id = self.alias_to_job_id.get(alias)
        if job_id is None:
            return None
        return self.inference_jobs.get(job_id)

    def _unregister_job(self, job: InferenceJob) -> list[str]:
        aliases = [
            alias
            for alias, job_id in list(self.alias_to_job_id.items())
            if job_id == job.id
        ]
        for alias in aliases:
            self.alias_to_job_id.pop(alias, None)
        self.inference_jobs.pop(job.id, None)
        return aliases

    def launch_from_request(
        self,
        launch_request: InferenceLaunchRequest,
        trainer_controller: Optional[TrainerControlServer] = None,
        *,
        reuse_existing: bool = True,
        preallocated_gpu_ids: Optional[list[int]] = None,
    ) -> InferenceJob:
        model_name = launch_request.model

        if reuse_existing and launch_request.owner is None:
            existing_job = self._get_job_by_alias(model_name)
            if existing_job is not None:
                self.logger.debug(
                    "Reusing existing inference job %s for model %s",
                    existing_job.id,
                    model_name,
                )
                return existing_job

        launch_config = InferenceLaunchConfig(
            max_context_length=launch_request.max_context_length,
            gpu_ids=list(preallocated_gpu_ids) if preallocated_gpu_ids else None,
            num_gpus=launch_request.num_gpus,
            log_file_path=launch_request.log_file_path,
        )

        if preallocated_gpu_ids:
            launch_config.num_gpus = len(preallocated_gpu_ids)

        if launch_config.num_gpus is not None and launch_config.num_gpus <= 0:
            raise ValueError("num_gpus must be greater than 0")

        owner = launch_request.owner
        if owner is not None:
            if owner.type == "grpo":
                launch_config.grpo_job_id = owner.job_id
            else:
                raise ValueError(
                    f"Unsupported inference launch owner type: {owner.type}"
                )

        if launch_config.grpo_job_id and trainer_controller is None:
            raise ValueError("Trainer controller is required for GRPO inference jobs")

        return self.launch_job(model_name, launch_config, trainer_controller)

    def launch_job(
        self,
        model: str,
        launch_config: InferenceLaunchConfig,
        trainer_controller: Optional[TrainerControlServer] = None,
    ) -> InferenceJob:
        is_grpo_sub_job = bool(launch_config.grpo_job_id)
        inference_job = InferenceJob(self.config, is_grpo_sub_job=is_grpo_sub_job)

        # Use provided GPU IDs or allocate through GPU manager
        allocated_here = False
        if launch_config.gpu_ids is None:
            requested_gpus = launch_config.num_gpus or 1
            allocated_gpus = self.gpu_manager.allocate_gpus(
                inference_job.id, requested_gpus
            )
            launch_config.gpu_ids = allocated_gpus
            allocated_here = True
            self.logger.info(
                f"Allocated GPUs {allocated_gpus} for inference job {inference_job.id}"
            )

        assert launch_config.gpu_ids is not None, "GPU IDs must be set before launching"
        assert len(launch_config.gpu_ids) > 0, (
            "Inference Job must have at least one GPU in gpu_ids. "
            f"Currently set to {launch_config.gpu_ids}"
        )

        try:
            inference_job.launch(model, launch_config, trainer_controller)
        except Exception:
            if allocated_here:
                self.gpu_manager.release_gpus(inference_job.id)
            raise

        aliases = [inference_job.id]
        if launch_config.grpo_job_id:
            aliases.append(launch_config.grpo_job_id)
        else:
            aliases.append(model)

        self._register_job(inference_job, aliases)

        self.logger.debug(
            "Active inference job aliases: %s",
            list(self.alias_to_job_id.keys()),
        )
        return inference_job

    def cleanup(self) -> None:
        """Clean up all inference jobs and their resources"""
        if self._cleanup_called and not self.inference_jobs:
            return

        self.logger.info(f"Cleaning up {len(self.inference_jobs)} inference jobs...")

        for job_id in list(self.inference_jobs.keys()):
            try:
                self.terminate_job(job_id)
            except Exception as e:
                self.logger.error(f"Error cleaning up inference job {job_id}: {e}")
        self._cleanup_called = True
        self.logger.info("InferenceManager cleanup completed")

    def terminate_job(
        self,
        job_id: str,
    ) -> dict[str, Any]:
        job = self.inference_jobs.get(job_id)
        if job is None:
            raise ValueError(f"Inference job '{job_id}' not found")

        aliases: list[str] = []
        release_exception: Optional[Exception] = None
        try:
            try:
                self.gpu_manager.release_gpus(job.id)
            except Exception as exc:
                release_exception = exc
                self.logger.error(
                    "Failed to release GPUs for inference job %s: %s",
                    job.id,
                    exc,
                )

            job.terminate()
        finally:
            aliases = self._unregister_job(job)

        self.logger.info("Terminated inference job %s (aliases=%s)", job.id, aliases)

        if release_exception is not None:
            raise release_exception

        return {"job_id": job.id, "aliases": aliases}



from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from arbor.core.logging import get_logger
from arbor.server.api.routes import grpo, inference

logger = get_logger(__name__)


# Small helper to serialize exceptions for responses/logging
def make_error_payload(exc: Exception) -> dict[str, Any]:
    return {
        "type": type(exc).__name__,
        "message": str(exc),
        "timestamp": datetime.now().isoformat(),
    }


def cleanup_managers(app: FastAPI):
    """Clean up all managers and their resources"""
    logger.info("Starting application cleanup...")

    # List of managers that should have cleanup methods
    manager_names = [
        "gpu_manager",
        "job_manager",
        "inference_manager",
        "grpo_manager",
        "file_train_manager",
        "file_manager",
    ]

    for manager_name in manager_names:
        if hasattr(app.state, manager_name):
            manager = getattr(app.state, manager_name)
            if hasattr(manager, "cleanup"):
                try:
                    logger.info(f"Cleaning up {manager_name}...")
                    manager.cleanup()
                except Exception as e:
                    logger.error(f"Error cleaning up {manager_name}: {e}")
        else:
            logger.debug(f"No {manager_name} found in app state")

    logger.info("Application cleanup completed")


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Lifespan event handler for FastAPI app."""
    # Startup
    logger.info("Application startup completed")

    yield

    # Shutdown
    logger.info("Application shutdown initiated")
    cleanup_managers(app)


app = FastAPI(title="Arbor API", lifespan=lifespan)


# Include routers
app.include_router(grpo.router, prefix="/v1/fine_tuning/grpo")
app.include_router(inference.router, prefix="/v1/chat")


# ---- Centralized exception handling ----


@app.exception_handler(Exception)
async def unhandled_exception_handler(_: Request, exc: Exception):
    # Map common exception types to HTTP statuses; default to 500
    status_map = {
        ValueError: 400,
        KeyError: 400,
        FileNotFoundError: 404,
        TimeoutError: 504,
        ConnectionError: 502,
    }
    status = 500
    for etype, code in status_map.items():
        if isinstance(exc, etype):
            status = code
            break
    return JSONResponse(status_code=status, content=make_error_payload(exc))



import json
import time
from datetime import datetime
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import dspy
import openai
import requests
from dspy.clients.provider import Provider, ReinforceJob, TrainingJob
from dspy.clients.utils_finetune import (
    GRPOGroup,
    GRPOStatus,
    TrainDataFormat,
    TrainingStatus,
    save_data,
)

if TYPE_CHECKING:
    from dspy.clients.lm import LM


class ArborTrainingJob(TrainingJob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_file_id = None
        self.provider_job_id = None

    def cancel(self):
        if ArborProvider.does_job_exist(self.provider_job_id):
            status = self.status()
            if ArborProvider.is_terminal_training_status(status):
                err_msg = "Jobs that are complete cannot be canceled."
                err_msg += f" Job with ID {self.provider_job_id} is done."
                raise Exception(err_msg)
            openai.fine_tuning.jobs.cancel(self.provider_job_id)
            self.provider_job_id = None

        if self.provider_file_id is not None:
            if ArborProvider.does_file_exist(self.provider_file_id):
                openai.files.delete(self.provider_file_id)
            self.provider_file_id = None

        super().cancel()

    def status(self) -> TrainingStatus:
        status = ArborProvider.get_training_status(self.provider_job_id)
        return status


class ArborReinforceJob(ReinforceJob):
    DEFAULT_TRAIN_KWARGS = {
        "temperature": 1.0,
        "beta": 0.00,
        "num_iterations": 1,
        "per_device_train_batch_size": 8,
        "learning_rate": 1e-6,
        "gradient_accumulation_steps": 1,
        # This is false by default in TRL, but I think it makes sense to be true for us
        "gradient_checkpointing": True,
        "lr_scheduler_type": "constant_with_warmup",
        "warmup_steps": 10,
        "max_prompt_length": None,
        "max_completion_length": None,
        "gradient_checkpointing_kwargs": None,
        "bf16": False,
        "scale_rewards": True,
        "max_grad_norm": 1.0,
        "report_to": "none",
        "log_completions": True,
        "logging_steps": 10,
        # By default, none is the model's max context length
        "max_context_length": None,
        "lora_config": None,
        "loss_type": "dapo",
        "soft_completion_penalty_length": None,
        "mask_truncated_completions": True,
        "weight_decay": 0.0,
    }

    def __init__(self, lm: "LM", train_kwargs: dict[str, Any]):
        # The teleprompter must ensure that this is set
        if "num_generations" not in train_kwargs:
            raise ValueError("num_generations must be set in the training kwargs")

        self.lm = lm
        self.train_kwargs = train_kwargs
        self.provider_job_id = None
        self.checkpoints: dict[str, dict[str, Any]] = {}
        self.last_checkpoint: str | None = None

    def initialize(self):
        # TODO(GRPO Team): Set provider job ID
        num_generations = self.train_kwargs.get("num_generations")
        temperature = self.train_kwargs.get(
            "temperature", self.DEFAULT_TRAIN_KWARGS["temperature"]
        )
        beta = self.train_kwargs.get("beta", self.DEFAULT_TRAIN_KWARGS["beta"])
        per_device_train_batch_size = self.train_kwargs.get(
            "per_device_train_batch_size",
            self.DEFAULT_TRAIN_KWARGS["per_device_train_batch_size"],
        )
        learning_rate = self.train_kwargs.get(
            "learning_rate", self.DEFAULT_TRAIN_KWARGS["learning_rate"]
        )
        gradient_accumulation_steps = self.train_kwargs.get(
            "gradient_accumulation_steps",
            self.DEFAULT_TRAIN_KWARGS["gradient_accumulation_steps"],
        )
        gradient_checkpointing = self.train_kwargs.get(
            "gradient_checkpointing",
            self.DEFAULT_TRAIN_KWARGS["gradient_checkpointing"],
        )
        lr_scheduler_type = self.train_kwargs.get(
            "lr_scheduler_type", self.DEFAULT_TRAIN_KWARGS["lr_scheduler_type"]
        )
        warmup_steps = self.train_kwargs.get(
            "warmup_steps", self.DEFAULT_TRAIN_KWARGS["warmup_steps"]
        )
        loss_type = self.train_kwargs.get(
            "loss_type", self.DEFAULT_TRAIN_KWARGS["loss_type"]
        )
        max_prompt_length = self.train_kwargs.get(
            "max_prompt_length", self.DEFAULT_TRAIN_KWARGS["max_prompt_length"]
        )
        max_completion_length = self.train_kwargs.get(
            "max_completion_length", self.DEFAULT_TRAIN_KWARGS["max_completion_length"]
        )
        soft_completion_penalty_length = self.train_kwargs.get(
            "soft_completion_penalty_length",
            self.DEFAULT_TRAIN_KWARGS["soft_completion_penalty_length"],
        )
        mask_truncated_completions = self.train_kwargs.get(
            "mask_truncated_completions",
            self.DEFAULT_TRAIN_KWARGS["mask_truncated_completions"],
        )
        bf16 = self.train_kwargs.get("bf16", self.DEFAULT_TRAIN_KWARGS["bf16"])
        scale_rewards = self.train_kwargs.get(
            "scale_rewards", self.DEFAULT_TRAIN_KWARGS["scale_rewards"]
        )
        gradient_checkpointing_kwargs = self.train_kwargs.get(
            "gradient_checkpointing_kwargs",
            self.DEFAULT_TRAIN_KWARGS["gradient_checkpointing_kwargs"],
        )
        max_grad_norm = self.train_kwargs.get(
            "max_grad_norm", self.DEFAULT_TRAIN_KWARGS["max_grad_norm"]
        )
        report_to = self.train_kwargs.get(
            "report_to", self.DEFAULT_TRAIN_KWARGS["report_to"]
        )
        log_completions = self.train_kwargs.get(
            "log_completions", self.DEFAULT_TRAIN_KWARGS["log_completions"]
        )
        logging_steps = self.train_kwargs.get(
            "logging_steps", self.DEFAULT_TRAIN_KWARGS["logging_steps"]
        )
        max_context_length = self.train_kwargs.get(
            "max_context_length", self.DEFAULT_TRAIN_KWARGS["max_context_length"]
        )
        max_steps = self.train_kwargs.get("max_steps", 500)
        num_training_gpus = self.train_kwargs.get("num_training_gpus", 1)
        num_inference_gpus = self.train_kwargs.get("num_inference_gpus", 1)
        lora_config = self.train_kwargs.get(
            "lora_config", self.DEFAULT_TRAIN_KWARGS["lora_config"]
        )
        weight_decay = self.train_kwargs.get(
            "weight_decay", self.DEFAULT_TRAIN_KWARGS["weight_decay"]
        )
        api_base = self.lm.kwargs["api_base"]

        finetune_model = ArborProvider._remove_provider_prefix(self.lm.model)
        data = {
            "model": finetune_model,
            "trainer_config": {
                "num_generations": num_generations,
                "temperature": temperature,
                "beta": beta,
                "per_device_train_batch_size": per_device_train_batch_size,
                "learning_rate": learning_rate,
                "gradient_accumulation_steps": gradient_accumulation_steps,
                "gradient_checkpointing": gradient_checkpointing,
                "lr_scheduler_type": lr_scheduler_type,
                "warmup_steps": warmup_steps,
                "max_prompt_length": max_prompt_length,
                "max_completion_length": max_completion_length,
                "soft_completion_penalty_length": soft_completion_penalty_length,
                "mask_truncated_completions": mask_truncated_completions,
                "bf16": bf16,
                "scale_rewards": scale_rewards,
                "gradient_checkpointing_kwargs": gradient_checkpointing_kwargs,
                "max_grad_norm": max_grad_norm,
                "report_to": report_to,
                "log_completions": log_completions,
                "logging_steps": logging_steps,
                # "max_context_length": max_context_length,
                # "max_seq_len": max_context_length,
                "max_steps": max_steps,
                "loss_type": loss_type,
                "num_training_gpus": num_training_gpus,
                "lora_config": lora_config,
                "weight_decay": weight_decay,
            },
            "inference_config": {
                "model": finetune_model,
                "max_context_length": max_context_length,
            },
            "gpu_config": {
                "type": "multi",
                "multi": {
                    "num_inference_gpus": num_inference_gpus,
                    "num_training_gpus": num_training_gpus,
                },
            },
        }
        url = urljoin(api_base, "fine_tuning/grpo/initialize")
        headers = {"Content-Type": "application/json"}
        response = requests.post(url=url, headers=headers, json=data)
        print(json.dumps(response.json(), indent=2))
        response.raise_for_status()
        response = response.json()
        self.lm.model = ArborProvider._add_provider_prefix(response["current_model"])
        self.provider_job_id = response.get("job_id")

    def _run_grpo_step_one_group(
        self,
        train_group: GRPOGroup,
        train_data_format: TrainDataFormat | str | None = None,
    ):
        # TODO: Check that the data follows the intended format
        api_base = self.lm.kwargs["api_base"]
        # api_key = self.lm.kwargs["api_key"]

        finetune_model = ArborProvider._remove_provider_prefix(self.lm.model)
        data = {
            "job_id": self.provider_job_id,
            "model": finetune_model,
            "batch": train_group["group"],
            "batch_id": train_group["batch_id"],
        }
        if "metrics" in train_group:
            data["metrics"] = train_group["metrics"] or {}
        url = urljoin(api_base, "fine_tuning/grpo/step")
        headers = {"Content-Type": "application/json"}
        response = requests.post(url, headers=headers, json=data)
        assert response.status_code == 200, (
            f"Failed to run a GRPO step: {response.text}"
        )
        response = response.json()
        assert "current_model" in response, (
            f"Response does not contain the next model ID to be used: {response}"
        )
        current_model = response["current_model"]
        self.lm.model = ArborProvider._add_provider_prefix(current_model)

    def step(
        self,
        train_data: list[GRPOGroup],
        train_data_format: TrainDataFormat | str | None,
    ):
        # Note: TrainDataFormat specifies the format for the inner most dict.
        # Because we run GRPO at the group level, train_data will be a list of
        # groups, where each group is a list of GRPOChatData. Our teleprompters
        # ensure that we pass the right data format.
        # We can consider making this distinction clearer, e.g., by having two
        # different step methods or changing our smallets data format to be the
        # GRPO group.
        # TODO: Support step on the server side
        assert train_data_format == TrainDataFormat.GRPO_CHAT, (
            f"GRPO only supports the GRPO_CHAT data format. Got {train_data_format} instead."
        )
        for group in train_data:
            self._run_grpo_step_one_group(group, train_data_format)

    def get_status(self) -> GRPOStatus:
        api_base = self.lm.kwargs["api_base"]
        url = urljoin(api_base, "fine_tuning/grpo/status")
        headers = {"Content-Type": "application/json"}
        body = {"job_id": self.provider_job_id}
        response = requests.post(url, headers=headers, json=body)
        assert response.status_code == 200, (
            f"Failed to get GRPO status: {response.text}"
        )
        return GRPOStatus(**response.json())

    def save_checkpoint(
        self,
        checkpoint_name: str,
        score: float | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        api_base = self.lm.kwargs["api_base"]
        url = urljoin(api_base, "fine_tuning/grpo/checkpoint")
        headers = {"Content-Type": "application/json"}
        body = {"checkpoint_name": checkpoint_name, "job_id": self.provider_job_id}
        if metadata is not None:
            body["metadata"] = metadata
        response = requests.post(url, headers=headers, json=body)
        response.raise_for_status()
        data = response.json()
        checkpoints = data.get("checkpoints", {})
        last_checkpoint = data.get("last_checkpoint")
        if not last_checkpoint:
            raise RuntimeError("Checkpoint request did not provide a checkpoint name")
        checkpoint = checkpoints.get(last_checkpoint, {}) or {
            "checkpoint_name": last_checkpoint
        }
        if score is not None:
            checkpoint["score"] = score
        if metadata is not None:
            checkpoint.setdefault("metadata", metadata)
        self.checkpoints[last_checkpoint] = checkpoint
        self.last_checkpoint = last_checkpoint

    def terminate(self):
        api_base = self.lm.kwargs["api_base"]

        url = urljoin(api_base, "fine_tuning/grpo/terminate")
        headers = {"Content-Type": "application/json"}
        body = {"job_id": self.provider_job_id}
        response = requests.post(url, headers=headers, json=body)
        assert response.status_code == 200, f"Failed to terminate GRPO: {response.text}"

        response = response.json()
        current_model = response["current_model"]
        self.lm.model = ArborProvider._add_provider_prefix(current_model)

    def cancel(self):
        if ArborProvider.does_job_exist(self.provider_job_id):
            status = self.status()
            if ArborProvider.is_terminal_training_status(status):
                err_msg = "Jobs that are complete cannot be canceled."
                err_msg += f" Job with ID {self.provider_job_id} is done."
                raise Exception(err_msg)
            openai.fine_tuning.jobs.cancel(self.provider_job_id)
            self.provider_job_id = None

    def status(self) -> TrainingStatus:
        status = ArborProvider.get_training_status(self.provider_job_id)
        return status


class ArborProvider(Provider):
    def __init__(self):
        super().__init__()
        self.finetunable = True
        self.reinforceable = True
        self.TrainingJob = ArborTrainingJob
        self.ReinforceJob = ArborReinforceJob

    @staticmethod
    def launch(lm: "LM", launch_kwargs: dict[str, Any] | None = None):
        model = ArborProvider._remove_provider_prefix(lm.model)

        api_base = lm.kwargs["api_base"]

        launch_kwargs = launch_kwargs or lm.launch_kwargs

        # Make request to launch endpoint
        response = requests.post(
            urljoin(api_base, "chat/launch"),
            json={"model": model, "launch_kwargs": launch_kwargs},
        )

        if response.status_code != 200:
            raise Exception(
                f"Failed to launch model. Status code: {response.status_code}, Response: {response.text}"
            )

        print(f"Inference server for model {model} launched successfully")

    @staticmethod
    def kill(lm: "LM", launch_kwargs: dict[str, Any] | None = None):
        api_base = lm.kwargs["api_base"]

        response = requests.post(
            urljoin(api_base, "chat/kill"),
        )

        if response.status_code != 200:
            raise Exception(
                f"Failed to kill model. Status code: {response.status_code}, Response: {response.text}"
            )

        print("Inference killed successfully")

    @staticmethod
    def _remove_provider_prefix(model: str) -> str:
        if model.startswith("openai/"):
            model = model[7:]
        if model.startswith("arbor:"):
            model = model[6:]
        return model

    @staticmethod
    def _add_provider_prefix(model: str) -> str:
        if not model.startswith("openai/arbor:"):
            model = "openai/arbor:" + model
        return model

    @staticmethod
    def _get_arbor_base_api():
        # TODO: We will delete this method once we start passing the LM object
        # to finetune.
        import dspy.settings as settings

        if not hasattr(settings, "arbor_api_base"):
            raise ValueError(
                "Arbor API base not set. Please set the `dspy.settings.arbor_api_base` to the URL for the Arbor server (e.g. 'http://localhost:8000/v1/')."
            )
        return dspy.settings.arbor_api_base

    @staticmethod
    def finetune(
        job: ArborTrainingJob,
        model: str,
        train_data: list[dict[str, Any]],
        train_data_format: TrainDataFormat | None = TrainDataFormat.CHAT,
        train_kwargs: dict[str, Any] | None = {},
    ) -> str:
        # TODO: We want to re-factor finetune so that it takes in an LM.
        # Until then, we use the following to get the api information. The
        # following is a dummy call to ensure that dspy.settings.arbor_base_api
        # is set.
        ArborProvider._get_arbor_base_api()

        model = ArborProvider._remove_provider_prefix(model)

        print("[Arbor Provider] Validating the data format")
        ArborProvider.validate_data_format(train_data_format)

        print("[Arbor Provider] Saving the data to a file")
        data_path = save_data(train_data)
        print(f"[Arbor Provider] Data saved to {data_path}")

        print("[Arbor Provider] Uploading the data to the provider")
        provider_file_id = ArborProvider.upload_data(data_path)
        job.provider_file_id = provider_file_id

        print("[Arbor Provider] Starting remote training")
        provider_job_id = ArborProvider._start_remote_training(
            train_file_id=job.provider_file_id,
            model=model,
            train_kwargs=train_kwargs or {},
        )
        job.provider_job_id = provider_job_id
        print(f"[Arbor Provider] Job started with the Arbor Job ID {provider_job_id}")

        print("[Arbor Provider] Waiting for training to complete")
        ArborProvider.wait_for_job(job)

        print("[Arbor Provider] Attempting to retrieve the trained model")
        model = ArborProvider.get_trained_model(job)
        print(f"[Arbor Provider] Model retrieved: {model}")

        return ArborProvider._add_provider_prefix(model)

    @staticmethod
    def does_job_exist(job_id: str) -> bool:
        try:
            original_base_url = openai.base_url
            openai.base_url = ArborProvider._get_arbor_base_api()
            openai.fine_tuning.jobs.retrieve(job_id)
            openai.base_url = original_base_url
            return True
        except Exception:
            return False

    @staticmethod
    def does_file_exist(file_id: str) -> bool:
        try:
            original_base_url = openai.base_url
            openai.base_url = ArborProvider._get_arbor_base_api()
            openai.files.retrieve(file_id)
            openai.base_url = original_base_url
            return True
        except Exception:
            return False

    @staticmethod
    def is_terminal_training_status(status: TrainingStatus) -> bool:
        return status in [
            TrainingStatus.succeeded,
            TrainingStatus.failed,
            TrainingStatus.cancelled,
        ]

    @staticmethod
    def get_training_status(job_id: str) -> TrainingStatus:
        provider_status_to_training_status = {
            "validating_files": TrainingStatus.pending,
            "queued": TrainingStatus.pending,
            "running": TrainingStatus.running,
            "succeeded": TrainingStatus.succeeded,
            "failed": TrainingStatus.failed,
            "cancelled": TrainingStatus.cancelled,
            "pending": TrainingStatus.pending,
            "pending_pause": TrainingStatus.pending,
            "pending_resume": TrainingStatus.pending,
            "paused": TrainingStatus.pending,
            "pending_cancel": TrainingStatus.pending,
        }

        if job_id is None:
            print("There is no active job.")
            return TrainingStatus.not_started

        err_msg = f"Job with ID {job_id} does not exist."
        assert ArborProvider.does_job_exist(job_id), err_msg

        original_base_url = openai.base_url
        openai.base_url = ArborProvider._get_arbor_base_api()
        provider_job = openai.fine_tuning.jobs.retrieve(job_id)
        openai.base_url = original_base_url

        provider_status = provider_job.status
        status = provider_status_to_training_status[provider_status]

        return status

    @staticmethod
    def validate_data_format(data_format: TrainDataFormat):
        supported_data_formats = [
            TrainDataFormat.CHAT,
            TrainDataFormat.COMPLETION,
            TrainDataFormat.GRPO_CHAT,
        ]

        if data_format not in supported_data_formats:
            err_msg = f"Arbor does not support the data format {data_format}."
            raise ValueError(err_msg)

    @staticmethod
    def upload_data(data_path: str) -> str:
        original_base_url = openai.base_url
        openai.base_url = ArborProvider._get_arbor_base_api()
        provider_file = openai.files.create(
            file=open(data_path, "rb"),
            purpose="fine-tune",
        )
        openai.base_url = original_base_url

        return provider_file.id

    @staticmethod
    def _start_remote_training(
        train_file_id: str, model: str, train_kwargs: dict[str, Any]
    ) -> str:
        original_base_url = openai.base_url
        openai.base_url = ArborProvider._get_arbor_base_api()
        provider_job = openai.fine_tuning.jobs.create(
            model=model,
            training_file=train_file_id,
            hyperparameters=train_kwargs,
        )
        openai.base_url = original_base_url
        return provider_job.id

    @staticmethod
    def wait_for_job(
        job: TrainingJob,
        poll_frequency: int = 20,
    ):
        done = False
        cur_event_id = None
        reported_estimated_time = False
        while not done:
            # Report estimated time if not already reported
            if not reported_estimated_time:
                original_base_url = openai.base_url
                openai.base_url = ArborProvider._get_arbor_base_api()
                remote_job = openai.fine_tuning.jobs.retrieve(job.provider_job_id)
                openai.base_url = original_base_url

                timestamp = remote_job.estimated_finish
                if timestamp:
                    estimated_finish_dt = datetime.fromtimestamp(timestamp)
                    delta_dt = estimated_finish_dt - datetime.now()
                    print(
                        f"[Arbor Provider] The Arbor estimated time remaining is: {delta_dt}"
                    )
                    reported_estimated_time = True

            # Get new events
            original_base_url = openai.base_url
            openai.base_url = ArborProvider._get_arbor_base_api()
            page = openai.fine_tuning.jobs.list_events(
                fine_tuning_job_id=job.provider_job_id, limit=1
            )
            openai.base_url = original_base_url

            new_event = page.data[0] if page.data else None
            if new_event and new_event.id != cur_event_id:
                dt = datetime.fromtimestamp(new_event.created_at)
                print(f"[Arbor Provider] {dt} {new_event.message}")
                cur_event_id = new_event.id

            # Sleep and update the flag
            time.sleep(poll_frequency)
            done = ArborProvider.is_terminal_training_status(job.status())

    @staticmethod
    def get_trained_model(job):
        status = job.status()
        if status != TrainingStatus.succeeded:
            err_msg = f"Job status is {status}."
            err_msg += f" Must be {TrainingStatus.succeeded} to retrieve model."
            raise Exception(err_msg)

        original_base_url = openai.base_url
        openai.base_url = ArborProvider._get_arbor_base_api()
        provider_job = openai.fine_tuning.jobs.retrieve(job.provider_job_id)
        openai.base_url = original_base_url

        finetuned_model = provider_job.fine_tuned_model
        return finetuned_model



"""
Enhanced logging system for Arbor with debugging and observability focus.

This module provides structured logging with context, request tracing, and debugging utilities
designed to make development and troubleshooting as easy as possible.
"""

import json
import logging
import sys
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional


class ArborFormatter(logging.Formatter):
    """Enhanced formatter with colors, context, and structured data."""

    # ANSI color codes
    COLORS = {
        "DEBUG": "\033[36m",  # Cyan
        "INFO": "\033[32m",  # Green
        "WARNING": "\033[33m",  # Yellow
        "ERROR": "\033[31m",  # Red
        "CRITICAL": "\033[35m",  # Magenta
        "RESET": "\033[0m",  # Reset
        "BOLD": "\033[1m",  # Bold
        "DIM": "\033[2m",  # Dim
    }

    # Component name mappings for cleaner output
    NAME_MAPPINGS = {
        "arbor.server.services.managers.inference_manager": "infer",
        "arbor.server.services.managers.grpo_manager": "grpo",
        "arbor.server.services.managers.file_manager": "files",
        "arbor.server.services.managers.health_manager": "health",
        "arbor.server.services.managers.job_manager": "jobs",
        "arbor.server.services.managers.file_train_manager": "train",
        "arbor.training.sft.trainer": "sft",
        "arbor.training.grpo.trainer": "grpo",
        "arbor.server.services.inference.vllm_client": "vllm",
        "arbor.server.services.inference.vllm_serve": "vllm-srv",
        "arbor.server.services.jobs.inference_job": "inf-job",
        "arbor.server.services.jobs.grpo_job": "grpo-job",
        "arbor.server.services.jobs.file_train_job": "train-job",
        "arbor.server.api.routes.inference": "api-inf",
        "arbor.server.api.routes.grpo": "api-grpo",
        "arbor.server.api.routes.files": "api-files",
        "arbor.server.api.routes.jobs": "api-jobs",
        "arbor.config": "config",
        "arbor.cli": "cli",
        "__main__": "main",
        "uvicorn": "api",
        "uvicorn.access": "api-acc",
        "uvicorn.error": "api-err",
        "fastapi": "api",
    }

    def __init__(self, show_context: bool = True, show_colors: bool = True):
        super().__init__()
        self.show_context = show_context
        self.show_colors = show_colors and sys.stderr.isatty()

    def format(self, record: logging.LogRecord) -> str:
        # Parse structured message if present
        message = record.getMessage()
        context_data = {}

        if " | {" in message and message.endswith("}"):
            try:
                msg_part, json_part = message.rsplit(" | ", 1)
                context_data = json.loads(json_part)
                message = msg_part
            except (ValueError, json.JSONDecodeError):
                pass  # Not structured, use as-is

        # Get short component name
        component = self.NAME_MAPPINGS.get(record.name, record.name)
        if len(component) > 10:
            component = component[:10]

        # Format timestamp
        timestamp = self.formatTime(record, "%H:%M:%S")

        # Apply colors if enabled
        if self.show_colors:
            level_color = self.COLORS.get(record.levelname, "")
            reset = self.COLORS["RESET"]
            bold = self.COLORS["BOLD"]
            dim = self.COLORS["DIM"]  # noqa: F841
            level = f"{level_color}{record.levelname[:4]}{reset}"
            component = f"{bold}{component}{reset}"
        else:
            level = record.levelname[:4]
        parts = [timestamp, f"[{level}]", f"[{component}]", message]

        # Add structured data as separate lines in debug mode
        if record.levelno == logging.DEBUG and context_data:
            formatted_context = json.dumps(context_data, indent=2)
            parts.append(f"\n  Context: {formatted_context}")

        return " ".join(parts)


class ArborLogger:
    """Enhanced logger with structured logging and context support."""

    def __init__(self, name: str):
        self.name = name
        # Get short name for cleaner output
        self.short_name = ArborFormatter.NAME_MAPPINGS.get(name, name)
        self._logger = logging.getLogger(self.short_name)

    def _log_with_context(
        self,
        level: int,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        exc_info: Optional[bool] = None,
        **kwargs,
    ):
        """Log with context and structured data."""
        if args:
            try:
                message = message % args
            except TypeError:
                message = message.format(*args)
        # Merge context and kwargs
        full_context: dict[str, Any] = {}
        if context is not None:
            if isinstance(context, Mapping):
                full_context.update(context)
            else:
                full_context["context"] = context
        if kwargs:
            full_context.update(kwargs)

        # Build structured message
        if full_context:
            structured_message = (
                f"{message} | {json.dumps(full_context, separators=(',', ':'))}"
            )
        else:
            structured_message = message

        self._logger.log(level, structured_message, exc_info=exc_info)

    def debug(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """Log debug message with context."""
        self._log_with_context(logging.DEBUG, message, *args, context=context, **kwargs)

    def info(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """Log info message with context."""
        self._log_with_context(logging.INFO, message, *args, context=context, **kwargs)

    def warning(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """Log warning message with context."""
        self._log_with_context(
            logging.WARNING, message, *args, context=context, **kwargs
        )

    def error(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        exc_info: bool = False,
        **kwargs,
    ):
        """Log error message with context."""
        self._log_with_context(
            logging.ERROR,
            message,
            *args,
            context=context,
            exc_info=exc_info,
            **kwargs,
        )

    def critical(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        exc_info: bool = False,
        **kwargs,
    ):
        """Log critical message with context."""
        self._log_with_context(
            logging.CRITICAL,
            message,
            *args,
            context=context,
            exc_info=exc_info,
            **kwargs,
        )

    def exception(
        self,
        message: str,
        *args,
        context: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """Log exception with context and stack trace."""
        self._log_with_context(
            logging.ERROR,
            message,
            *args,
            context=context,
            exc_info=True,
            **kwargs,
        )


def setup_logging(
    log_level: str = "INFO",
    log_dir: Optional[Path] = None,
    enable_file_logging: bool = True,
    enable_console_logging: bool = True,
    show_context: bool = True,
    show_colors: bool = True,
    debug_mode: bool = False,
) -> dict[str, Any]:
    """
    Setup enhanced logging for Arbor.

    Args:
        log_level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        log_dir: Directory for log files
        enable_file_logging: Whether to enable file logging
        enable_console_logging: Whether to enable console logging
        show_context: Whether to show context in console output
        show_colors: Whether to use colors in console output
        debug_mode: Enable debug mode with extra verbose logging

    Returns:
        Dictionary with logging configuration details
    """
    import os
    import tempfile

    # Use a persistent log directory during tests to avoid cleanup issues
    if "PYTEST_CURRENT_TEST" in os.environ or "pytest" in sys.modules:
        if log_dir is None or (log_dir and "/tmp" in str(log_dir)):
            # Use a stable temp directory that won't get cleaned up during tests
            stable_temp_dir = Path(tempfile.gettempdir()) / "arbor_test_logs"
            stable_temp_dir.mkdir(exist_ok=True)
            log_dir = stable_temp_dir

    # Set debug level if debug mode is enabled
    if debug_mode:
        log_level = "DEBUG"

    # Create formatters
    console_formatter = ArborFormatter(
        show_context=show_context, show_colors=show_colors
    )
    file_formatter = ArborFormatter(show_context=True, show_colors=False)

    # Get root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(getattr(logging, log_level.upper()))

    # Clear existing handlers
    root_logger.handlers.clear()

    handlers = []

    # Console handler
    if enable_console_logging:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(getattr(logging, log_level.upper()))
        console_handler.setFormatter(console_formatter)
        root_logger.addHandler(console_handler)
        handlers.append("console")

    # File handlers
    if enable_file_logging and log_dir:
        log_dir = Path(log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)

        # Main log file (all levels)
        main_log_file = log_dir / "arbor.log"
        file_handler = logging.FileHandler(main_log_file)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(file_formatter)
        root_logger.addHandler(file_handler)
        handlers.append("file")

        # Error log file (errors and critical only)
        error_log_file = log_dir / "arbor_error.log"
        error_handler = logging.FileHandler(error_log_file)
        error_handler.setLevel(logging.ERROR)
        error_handler.setFormatter(file_formatter)
        root_logger.addHandler(error_handler)
        handlers.append("error_file")

        # Debug log file (debug mode only)
        if debug_mode:
            debug_log_file = log_dir / "arbor_debug.log"
            debug_handler = logging.FileHandler(debug_log_file)
            debug_handler.setLevel(logging.DEBUG)
            debug_handler.setFormatter(file_formatter)
            root_logger.addHandler(debug_handler)
            handlers.append("debug_file")

    # Configure third-party loggers
    _configure_third_party_loggers(debug_mode)

    return {
        "level": log_level,
        "handlers": handlers,
        "log_dir": str(log_dir) if log_dir else None,
        "debug_mode": debug_mode,
        "show_context": show_context,
        "show_colors": show_colors,
    }


def _configure_third_party_loggers(debug_mode: bool = False):
    """Configure third-party library loggers."""

    # Set levels for third-party libraries
    third_party_levels = {
        "uvicorn": "INFO" if debug_mode else "WARNING",
        "uvicorn.access": "INFO" if debug_mode else "WARNING",
        "uvicorn.error": "INFO" if debug_mode else "WARNING",
        "fastapi": "INFO" if debug_mode else "WARNING",
        "httpx": "INFO" if debug_mode else "WARNING",
        "urllib3": "INFO" if debug_mode else "WARNING",
        "vllm": "INFO",
        "torch": "WARNING",
        "transformers": "WARNING",
        "accelerate": "WARNING",
        "datasets": "WARNING",
        "trl": "INFO" if debug_mode else "WARNING",
    }

    for logger_name, level in third_party_levels.items():
        logger = logging.getLogger(logger_name)
        logger.setLevel(getattr(logging, level))


def get_logger(name: str) -> ArborLogger:
    """
    Get an enhanced Arbor logger.

    Args:
        name: Logger name, typically __name__

    Returns:
        Enhanced ArborLogger instance
    """
    return ArborLogger(name)


def log_configuration(config: dict[str, Any]):
    """Log configuration information."""
    logger = get_logger("arbor.config")
    logger.info("Configuration loaded", config=config)



from fastapi import APIRouter, Request

from arbor.core.logging import get_logger
from arbor.server.api.schemas import (
    InferenceLaunchRequest,
    InferenceTerminateRequest,
)
from arbor.server.services.managers.inference_manager import InferenceManager
from arbor.utils.helpers import strip_prefix

logger = get_logger(__name__)

router = APIRouter()


@router.post("/completions")
async def run_inference(
    request: Request,
):
    inference_manager: InferenceManager = request.app.state.inference_manager
    raw_json = await request.json()
    raw_json["model"] = strip_prefix(raw_json["model"])

    # forward the request to the inference server
    completion = await inference_manager.route_inference(raw_json)

    # # Write combined request+completion to a single JSONL for inspection
    # try:
    #     combined = {
    #         "request": raw_json,
    #         "completion": completion,
    #     }
    #     with open("completion_inspect.jsonl", "a") as f:
    #         f.write(json.dumps(combined) + "\n")
    # except Exception as exc:
    #     logger.warning(f"Failed to write completion_inspect.jsonl: {exc}")

    return completion


@router.post("/launch")
async def launch_inference(request: Request, launch_request: InferenceLaunchRequest):
    inference_manager: InferenceManager = request.app.state.inference_manager
    normalized_request = launch_request.model_copy(
        # stripe the prefix (such as "openai/arbor:")
        update={"model": strip_prefix(launch_request.model)}
    )

    inference_job = inference_manager.launch_from_request(normalized_request)

    return {
        "message": "Inference server launched",
        "job_id": inference_job.id,
        "model": inference_job.launched_model_name,
    }


@router.post("/kill")
async def kill_inference(
    request: Request, terminate_request: InferenceTerminateRequest
):
    inference_manager: InferenceManager = request.app.state.inference_manager

    job_id = terminate_request.job_id

    if job_id == "*":
        inference_manager.cleanup()
        return {"message": "All inference servers terminated"}
    else:
        summary = inference_manager.terminate_job(
            job_id,
        )
        return {"message": "Inference server terminated", **summary}



"""DSPy integration for Arbor.

Exports Arbor's DSPy provider and GRPO trainer in a structured namespace.
"""

from .arbor_provider import ArborProvider
from .grpo_optimizer import ArborGRPO

__all__ = ["ArborGRPO", "ArborProvider"]
